/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2007  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef SERVER_H_
#define SERVER_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "AbstractServer.h"
#include "HttpResponseParser.h"
#include "ChainAsyncConnector.h"
#include "ChainConnectorListener.h"
#include "SplittableBuffer.h"
#include "ServerTimeouts.h"
#include "SymbolicInetAddr.h"
#include "AutoClosingSAP.h"
#include "HttpVersion.h"
#include "Reactor.h"
#include "EventHandler.h"
#include "NullRefCounter.h"
#include "IntrusivePtr.h"
#include "RequestPtr.h"
#include "WeakPtr.h"
#include "StrongPtr.h"
#include "NonCopyable.h"
#include "Timer.h"
#include <sigc++/sigc++.h>
#include <stddef.h>
#include <string>
#include <deque>
#include <memory>

class ServiceContext;
class ServerConnection;
class HttpRequestMetadata;
class HttpResponseMetadata;
class AbstractRequestHandler;
class AbstractResponseHandler;
class ErrorDescriptor;
class FilterTryList;
class RequestTag;
class AsyncConnectorError;
class SocksError;
class ACE_SOCK_Stream;

class Server :
	public AbstractServer,
	private HttpResponseParser,
	private ChainConnectorListener,
	private EventHandler<NullRefCounter>,
	public sigc::trackable
{
	DECLARE_NON_COPYABLE(Server)
public:
	Server(ServiceContext& context, ServerTimeouts const& timeouts);
	
	virtual ~Server();

	virtual IntrusivePtr<AbstractRequestHandler> submitRequest(
		ConstRequestPtr const& request_metadata,
		RequestTag const& request_tag,
		IntrusivePtr<AbstractResponseHandler> const& handler);
	
	virtual IntrusivePtr<AbstractRequestHandler> submitRequest(
		ConstRequestPtr const& request_metadata,
		RequestTag const& request_tag,
		IntrusivePtr<AbstractResponseHandler> const& handler,
		std::auto_ptr<ImmediateResponse> response);
	
	virtual bool isIdle() const;
	
	virtual bool isOverloaded() const;
	
	void reset();
private:
	enum { MAX_PIPELINE_SIZE = 4 };
	enum { READ_BUF_SIZE = 16*1024 };
	enum { MAX_BUFFERED_REQUEST_BODY = 1024*1024 };
	enum { MAX_WRITE_BUF_SIZE = 24000 };
	
	enum Flags {
		READING_SUSPENDED = 1,
		WRITE_BUFFER_FULL = 2
	};
	
	class Request;
	friend class Server::Request;
	class PhantomRequest;
	friend class Server::PhantomRequest;
	class ScopedSuspender;
	friend class Server::ScopedSuspender;
	class RequestCanceller;
	
	typedef WeakPtr<ACE_NULL_SYNCH, Server, Server*> ServerWeakPtr;
	typedef std::deque<IntrusivePtr<Request> > RequestQueue;
	typedef std::deque<IntrusivePtr<PhantomRequest> > PhantomRequestQueue;
	
	void enqueueRequest(IntrusivePtr<Request> const& req);
	
	void enqueuePhantomRequest(IntrusivePtr<PhantomRequest> const& req);
	
	void updatePipeline();
	
	static bool validateRequest(Request& req);
	
	static bool validateRequestProtocol(Request& req);
	
	static bool validateRequestExpectations(Request& req);
	
	static bool validateTunnelPort(Request& req);
	
	bool isNewConnectionRequiredFor(Request const& req) const;
	
	void requestConnectionFor(Request const& req);
	
	bool isOutputReady() const;
	
	size_t getPendingOutputSize() const;
	
	void onNewOutput();
	
	void processOutputData();
	
	size_t writeRequestData(Request& req);
	
	void processInputData(bool eof);
	
	void dbgReportOutgoingRequestSent();
	
	void shiftRequestQueue();
	
	bool connectionMayBeStored() const;
	
	void storeConnection();
	
	void dropConnection();
	
	std::auto_ptr<ServerConnection> retrieveConnection();
	
	void restartPendingRequests();
	
	void answerWithError(std::auto_ptr<ErrorDescriptor> edesc);
	
	bool isRequestAtPipelineStart(Request const* req) const;
	
	void suspend() { m_flags |= READING_SUSPENDED; }
	
	void resume() { m_flags &= ~READING_SUSPENDED; }
	
	void resetFlags() { m_flags &= READING_SUSPENDED; }
	
	void onReadTimeout();
	
	void onWriteTimeout();
	
	void onTimeout();
	
	void updateState();
	
	bool processLeadingPhantomResponse();
	
	bool processLeadingErrorResponse();
	
	void processImmediateResponses();
	
	bool isImmediateResponseRequired() const;
	
	virtual void eventGotProvisionalResponse(
		std::auto_ptr<HttpResponseMetadata> metadata);
	
	virtual void eventGotMetadata(
		std::auto_ptr<HttpResponseMetadata> metadata, bool is_persistent);
	
	virtual void onConnectionEstablished(
		AutoClosingSAP<ACE_SOCK_Stream>& conn);
	
	virtual void onConnectionFailed(
		AsyncConnectorError const& err, SymbolicInetAddr const& addr);
	
	virtual void onConnectionFailed(
		SocksError const& err, ProxyDescriptor const& proxy);
	
	virtual void onConnectionFailed(
		std::string const& error_message, ProxyDescriptor const& proxy);
	
	void handleForwardingProblem(
		std::string const& problem, ProxyDescriptor const& proxy);
	
	virtual void handleRead(ACE_HANDLE);
	
	virtual void handleWrite(ACE_HANDLE);
	
	void processNewHttpConnection(AutoClosingSAP<ACE_SOCK_Stream>& strm);
	
	void processNewHttpConnection(std::auto_ptr<ServerConnection> conn);
	
	std::auto_ptr<HttpResponseMetadata>
		createConnectionEstablishedResponse(Request const& req);
	
	void onDelayedAction();
	
	ServiceContext& m_rContext;
	ServerTimeouts m_timeouts;
	RequestQueue m_requestQueue;
	PhantomRequestQueue m_phantomRequestQueue;
	size_t m_pipelineSize;
	ChainAsyncConnector m_connector;
	std::auto_ptr<ServerConnection> m_ptrConnection;
	SplittableBuffer m_inputData;
	HttpVersion m_lastResponseVersion;
	ReactorHandlerId m_handlerId;
	Timer m_readTimer;
	Timer m_writeTimer;
	Timer m_delayedActionTimer;
	Reactor::IOEvents m_eventMask;
	int m_flags;
	int m_isWaitingForWrite;
	StrongPtr<ACE_NULL_SYNCH, Server, Server*> m_ptrSelf;
};

#endif
